/*
 ============================================================================
 Name        : main.c
 Author      : Maciej Poliwoda
 Version     :
 Copyright   : Your copyright notice
 Description : Hello World in C, Ansi-style
 ============================================================================
 */


#include <stdio.h>
#include <stdlib.h>
#include <isl/ctx.h>
#include <isl/set.h>
#include <isl/union_set.h>
#include <isl/flow.h>
#include <isl/space.h>
#include <isl/flow.h>
#include <pet.h>
#include "debug_tool.h"
#include "calc_scop.h"
#include "stencil_scop.h"
#include "str_func.h"
#include "codegen.h"
#include "isl_tool.h"
#include <isl/id.h>

int main(int argc, char *argv[] ) {
	pet_scop* pet;

	isl_ctx* ctx;
	calc_scop* calc;
	calc_scop_list* calc_list = 0;
	stencil_scop_list* stencil_list = 0;

	char* filename;
	char* paramValue;

	stencil_Param stencilParam;

	stencilParam.size = 16;
	stencilParam.sizesDimIndex = 0;
	stencilParam.sizesDim = 0;
	stencilParam.sizes = 0;
	stencilParam.method = 2;
	stencilParam.bound = 1000;

	if(is_param_set("--help", argc, argv)){
		char tabs[8];
		sprintf(tabs,"\t\t\t");
		printf("parameters\n");
		printf("\t%s%s%s\n", "filename=<path>  ", tabs, "file with loop to parse");
		printf("\t%s%s%s\n", "method=<int>     ", tabs, "method for tiles build");
		printf("\t%s%s%s\n", "[size=<int>]     ", tabs, "default size (default:16)");
		printf("\t%s%s%s\n", "[sizes=<int>]    ", tabs, "number of spaces with specified size (default:0)");
		printf("\t%s%s%s\n", "[size<int>=<int>]", tabs, "specifies size list for spaces, when sizes is set (default:empty list)");
		printf("\t%s%s%s\n", "[bound=<int>]    ", tabs, "specifies value of loop bounds for wavefront card calculations (default:1000)");

		printf("options\n");
		printf("\t%s%s%s\n", "--isl-schedule-on", tabs, "use isl schedule map for loop normalization");
		printf("\t%s%s%s\n", "--debug-print-on ", tabs, "print debug info to stderr");
		printf("\t%s%s%s\n", "--no-tiles       ", tabs, "only loop parse");
		printf("\t%s%s%s\n", "--help           ", tabs, "print this help, then exit");
		return EXIT_SUCCESS;
	}

	filename = get_param_value("filename=", argc, argv);
	if(!filename) return EXIT_SUCCESS;

	paramValue = get_param_value("size=", argc, argv);
	if(paramValue){
		stencilParam.size = atoi(paramValue);
	}

	paramValue = get_param_value("sizes=", argc, argv);
	if(paramValue){
		int i;
		char buf[32];
		stencilParam.sizesDim = atoi(paramValue);
		stencilParam.sizes = (int*)malloc(stencilParam.sizesDim*sizeof(int));
		for(i=0; i < stencilParam.sizesDim; i++ ){
			stencilParam.sizes[i] = stencilParam.size;
			sprintf(buf, "size%i=",i+1);
			paramValue = get_param_value(buf, argc, argv);
			if(paramValue){
				stencilParam.sizes[i] = atoi(paramValue);
			}
		}
	}


	paramValue = get_param_value("method=", argc, argv);
	if(paramValue){
		stencilParam.method = atoi(paramValue);
	}

	paramValue = get_param_value("bound=", argc, argv);
	if(paramValue){
		stencilParam.bound = atoi(paramValue);
	}

	if(is_param_set("--debug-print-on", argc, argv)){
		isl_debug_on();
	}

	if(!(stencilParam.method==1 || stencilParam.method==2 || stencilParam.method==3))  return EXIT_SUCCESS;

	ctx = isl_ctx_alloc_with_pet_options();
	pet_options_set_signed_overflow(ctx, PET_OVERFLOW_IGNORE);

	pet = pet_scop_extract_from_C_source(ctx, filename , NULL);

	if(!pet)
		return EXIT_SUCCESS;

	calc = calc_scop_from_pet_scop(ctx, pet);


	if(!is_param_set("--isl-schedule-on", argc, argv)){
		calc_list = calc_scop_normalize(calc, isl_bool_false );
	}
	else{
		calc_list = calc_scop_normalize(calc, isl_bool_true );
	}

	if(!is_param_set("--no-tiles", argc, argv))
		stencil_list = stencil_scop_list_from_calc_scop_list(ctx, calc_list, &stencilParam);

	printf("\noryginal loop:\n");

	isl_printf_union_set("\ndomain:\n%s\n", calc->domain);
	isl_printf_union_map("\nwrites:\n%s\n", calc->must_writes);
	isl_printf_union_map("\nreads:\n%s\n", calc->may_reads);
	isl_printf_schedule("\norginal schedule:\n%s\n", calc->schedule);


	isl_printf_union_map("\nglobal schedule:\n%s\n", calc->schedule_map);
	isl_printf_union_map("\nbefore:\n%s\n", calc->before);

	isl_printf_union_map("\ndep_raw:\n%s\n", calc->dep_raw);
	isl_printf_union_map("\ndep_waw:\n%s\n", calc->dep_waw);
	isl_printf_union_map("\ndep_war:\n%s\n", calc->dep_war);

	isl_printf_union_map("\nrelations:\n%s\n", calc->relations);

	isl_printf_union_set("\ndeltas relations:\n%s\n", calc->delta);

	isl_printf_schedule("\nisl schedule:\n%s\n", calc->islSchedule);
	isl_printf_union_map("\nisl schedule map:\n%s\n", calc->islScheduleMap);

	//isl_printf_schedule("\nisl schedule no tuple:\n%s\n", calc->islScheduleNoTuples);
	//isl_printf_union_map("\nisl schedule map no tuple:\n%s\n", calc->islScheduleMapNoTuples);

	{
		int i;

		for(i = 0; i < calc_list->counter; i++){
			calc_scop* calc_norm = calc_list->calc_scop_list[i];

			printf("\nnormalized loop (%d):\n", i);

			isl_printf_union_map("\nloop mapper:\n%s\n", calc_norm->schedule_mapper);
			isl_printf_union_set("\ndomain:\n%s\n", calc_norm->domain);
			isl_printf_union_map("\nwrites:\n%s\n", calc_norm->must_writes);
			isl_printf_union_map("\nreads:\n%s\n", calc_norm->may_reads);

			isl_printf_union_map("\nglobal schedule:\n%s\n", calc_norm->schedule_map);
			isl_printf_union_map("\nbefore:\n%s\n", calc_norm->before);

			isl_printf_union_map("\ndep_raw:\n%s\n", calc_norm->dep_raw);
			isl_printf_union_map("\ndep_waw:\n%s\n", calc_norm->dep_waw);
			isl_printf_union_map("\ndep_war:\n%s\n", calc_norm->dep_war);

			isl_printf_union_map("\nrelations:\n%s\n", calc_norm->relations);

			isl_printf_union_set("\ndeltas relations:\n%s\n", calc_norm->delta);

			isl_printf_schedule("\nisl schedule:\n%s\n", calc_norm->islSchedule);
			isl_printf_union_map("\nisl schedule map:\n%s\n", calc_norm->islScheduleMap);

			//isl_printf_schedule("\nisl schedule no tuple:\n%s\n", calc_norm->islScheduleNoTuples);
			//isl_printf_union_map("\nisl schedule map no tuple:\n%s\n", calc_norm->islScheduleMapNoTuples);

			if(!is_param_set("--no-tiles", argc, argv)){
				int j=0;
				stencil_scop* stencil = stencil_list->stencil_scop_list[i];

				isl_printf_union_set("\nall distance vectors:\n%s\n", stencil->deltaInfoSet);

				printf("\nloop dimensions:\n");
				for(j = 0; j < stencil->loopDim; j++){
					isl_printf_str("%s\n",stencil_scop_deltaDim_Info_to_str(stencil,j));
				}

				printf("\nspaces:\n");
				for(j = 0; j < stencil->spaceInfoDim; j++){
					isl_printf_str("%s\n",stencil_scop_space_Info_to_str(stencil,j));
				}

				isl_printf_union_set("\ntile:\n%s\n", stencil->tile);
				isl_printf_union_map("\ntile relations:\n%s\n", stencil->tileRelations);
				isl_printf_union_set("\ntile relations distance vectors:\n%s\n", stencil->tileRelationsDelta);
				isl_printf_union_map("\ntile schedule map:\n%s\n", stencil->tileScheduleMap);
				isl_printf_union_map("\nstencil global schedule:\n%s\n", stencil->stencilScheduleMap);
				isl_printf_union_map("\nstencil code schedule:\n%s\n", stencil->stencilGlobalSchedule);
				isl_printf_union_set("\nwavefront card unionSet:\n%s\n", stencil->stencilScheduleCardUnionSet);
				isl_printf_str("\nwavefront card value:\n%s\n", isl_val_to_str(stencil->wavefrontSetCardValue));
			}
		}
	}

	if(!is_param_set("--no-tiles", argc, argv))
	{
		int i;

		printf("\n###############################\n");

		isl_printf_union_map("\nstencil global code schedule:\n%s\n", stencil_list->stencilGlobalSchedule);

		printf("\n###############################\n");
		{
			isl_union_map *stencilBefore = isl_union_map_lex_lt_union_map(isl_union_map_copy(stencil_list->stencilGlobalSchedule),isl_union_map_copy(stencil_list->stencilGlobalSchedule));

			printf( "\nDoes stencil global code schedule respects oryginal loop RaW deps?\n");
			if(isl_union_map_is_subset(calc->dep_raw,stencilBefore) == isl_bool_true){
				printf("True\n");
			}
			else{
				printf("False\n");
			}
			printf( "\nDoes stencil global code schedule respects oryginal loop WaW deps?\n");
			if(isl_union_map_is_subset(calc->dep_waw, stencilBefore) == isl_bool_true){
				printf("True\n");
			}
			else{
				printf("False\n");
			}
			printf( "\nDoes stencil global code schedule respects oryginal loop WaR deps?\n");
			if(isl_union_map_is_subset(calc->dep_war, stencilBefore) == isl_bool_true){
				printf("True\n");
			}
			else{
				printf("False\n");
			}

			isl_union_map_free(stencilBefore);
		}
		printf("\n###############################\n");

		isl_printf_str("\nstencil codegen:\n%s", codegen_macros_to_str(stencil_list->stencilGlobalSchedule, pet));

		for(i = 0; i < calc_list->counter; i++){
			stencil_scop* stencil = stencil_list->stencil_scop_list[i];
			if(stencil->spaceInfoDim > 1 || isl_union_set_is_empty(stencil->tileRelationsDelta) == isl_bool_true )
			{
				isl_printf_str("%s", codegen_wavefront_to_str(stencil->stencilGlobalSchedule, pet, isl_bool_true));
			}
			else
			{
				isl_printf_str("%s", codegen_wavefront_to_str(stencil->stencilGlobalSchedule, pet, isl_bool_false));
			}
		}
	}

	if(is_param_set("--isl-schedule-show", argc, argv)){
		isl_union_map *schedule = isl_union_map_intersect_domain(isl_union_map_copy(calc->islScheduleMap), isl_union_set_copy(calc->domain));
		printf("\n###############################\n");
		printf("\nisl schedule code:\n");
		isl_printf_str("%s", codegen_wavefront_to_str(schedule, pet, isl_bool_false));
		printf("\n###############################\n");

		isl_union_map_free(schedule);
	}

	printf("\nEnd");

	pet_scop_free(pet);
	calc_scop_free(calc);
	calc_scop_list_free(calc_list);
	stencil_scop_list_free(stencil_list);
	isl_ctx_free(ctx);

	return EXIT_SUCCESS;
}


