/*
* stencil_scop.c
 *
 *  Created on: 22 pa 2018
 *      Author: Maciej
 */
#include "stencil_scop.h"
#include "str_func.h"
#include <string.h>
#include <isl/schedule_node.h>
#include <isl/constraint.h>
#include "isl_tool.h"
#include "debug_tool.h"

#define hyperplaneFrirstIndex 3

typedef struct dimInfo{
	int dimPos;
	stencil_scop* stencil;
} dimInfo;

static void init_info_in_stencil_scop(__isl_keep calc_scop* calc, stencil_scop* stencil);

static void delta_union_set_analyze(__isl_keep isl_union_set *delta_union_set, stencil_scop* stencil);
static isl_stat delta_set_analyze(__isl_take isl_set *set, void *user);
static isl_stat delta_set_point_analyze(__isl_take isl_point *pnt, void *user);

static void space_info_create_tile(__isl_keep isl_union_set *domain_union_set, stencil_scop* stencil);
static void space_info_create_hyperplane_for_dim(stencil_scop* stencil, int dim);
static isl_bool space_info_check_hyperplane(stencil_scop* stencil, __isl_keep isl_set *hyperplane);
static isl_stat space_info_create_hyperplane_for_delta_vector(__isl_take isl_point *pnt, void *user);
static void space_info_create_hyperplane_max_non_zero_dim(stencil_scop* stencil);
static void space_info_add_hyperplane_to_check_info(stencil_scop* stencil, __isl_keep isl_set *hyperplane);
static void space_info_hyperplane_build_equation(__isl_take isl_set* set, char* equationStr, int level, stencil_scop* stencil);
static isl_stat space_info_hyperplanes_dims_check(__isl_take isl_point *pnt, void *user);
static isl_stat space_info_hyperplanes_count(__isl_take isl_point *pnt, void *user);
static isl_stat space_info_from_hyperplanes(__isl_take isl_point *pnt, void *user);
static void space_info_add_space_to_set(stencil_scop* stencil, int spaceInfoPos);

static isl_stat space_info_build_tile_domain(__isl_take isl_set *set, void *user);
static void tile_relations_build_map(isl_union_map* relations, stencil_scop* stencil);
static isl_stat tile_relations_from_loop_relations(__isl_take isl_map *map, void *user);
static isl_stat tile_relations_from_map(__isl_take isl_map *map, void *user);

static void tile_relations_delta_stencil_schedule_create(__isl_keep isl_union_map* schedule_map, stencil_scop* stencil, int scopeDimSize);

static void tile_relations_delta_create(stencil_scop* stencil);
static isl_stat tile_relations_delta_set_to_vectors(__isl_take isl_set *set, void *user);
static isl_stat tile_relations_delta_set_point_analyze(__isl_take isl_point *pnt, void *user);
static __isl_give isl_set* tile_relations_delta_create_wafefront_hyperplane(stencil_scop* stencil);

static void stencil_global_schedule_create(__isl_keep isl_union_set *domain_union_set, __isl_keep isl_union_set *orig_domain_union_set, __isl_keep isl_union_map *mapper_union_map, stencil_scop* stencil);
static void stencil_global_schedule_scalar_create(__isl_keep isl_union_map *schedule_map, stencil_scop* stencil, calc_scop* calc);
static void stencil_global_schedule_card_calculate(stencil_scop* stencil);
static isl_stat stencil_global_schedule_card_set_build(__isl_take isl_set *set, void *user);
static isl_stat stencil_global_schedule_card_union_set_build(__isl_take isl_set *set, void *user);

__isl_give stencil_scop_list* stencil_scop_list_from_calc_scop_list(isl_ctx *ctx, __isl_keep calc_scop_list* calc_list, stencil_Param* stencilParam){
	int i;
	stencil_scop_list* stencil_list = (stencil_scop_list*)malloc(sizeof(stencil_scop_list));

	stencil_list->stencil_scop_list = (stencil_scop**)malloc(calc_list->counter*sizeof(stencil_scop*));
	stencil_list->counter = calc_list->counter;
	stencil_list->stencilGlobalSchedule = 0;

	for(i = 0; i < stencil_list->counter; i++){
		stencil_list->stencil_scop_list[i] = stencil_scop_from_calc_scop(ctx, calc_list->calc_scop_list[i], stencilParam, i);

		if(!stencil_list->stencilGlobalSchedule){
			stencil_list->stencilGlobalSchedule = isl_union_map_copy(stencil_list->stencil_scop_list[i]->stencilGlobalSchedule);
		}

		stencil_list->stencilGlobalSchedule = isl_union_map_coalesce(isl_union_map_union(stencil_list->stencilGlobalSchedule, isl_union_map_copy(stencil_list->stencil_scop_list[i]->stencilGlobalSchedule)));
	}

	return stencil_list;
}
__isl_null stencil_scop_list* stencil_scop_list_free(stencil_scop_list* stencil_list){
	int i = 0;

	if(!stencil_list) return 0;

	for(i = 0; i < stencil_list->counter; i++){
		stencil_scop_free(stencil_list->stencil_scop_list[i]);
	}
	free(stencil_list->stencil_scop_list);
	isl_union_map_free(stencil_list->stencilGlobalSchedule);
	free(stencil_list);

	return 0;
}


__isl_give stencil_scop* stencil_scop_from_calc_scop(isl_ctx *ctx, __isl_keep calc_scop* calc, stencil_Param* stencilParam, int stencil_index)
{
	stencil_scop* stencil;
	stencil = stencil_scop_alloc(ctx);

	stencil->stencil_index = stencil_index;

	isl_debug_printf("\n%s\n", "#stencil scop start");

	stencil->stencilParam = stencilParam;

	init_info_in_stencil_scop(calc, stencil);
	delta_union_set_analyze(calc->delta, stencil);

	space_info_create_tile(calc->domain, stencil);

	if(stencil->spaceInfoDim > 0){
		tile_relations_build_map(calc->relations, stencil);

		tile_relations_delta_stencil_schedule_create(calc->schedule_map, stencil, calc->scopeDimSize);

		stencil_global_schedule_create(calc->domain, calc->orig_domain, calc->schedule_mapper, stencil);

		stencil_global_schedule_card_calculate(stencil);
	}
	else{
		stencil_global_schedule_scalar_create(calc->schedule_map, stencil, calc);
	}

	isl_debug_printf("\n%s\n", "#stencil scop end");

	return stencil;
}

__isl_null stencil_scop *stencil_scop_free(__isl_take stencil_scop *stencil){
	if(!stencil)
			return NULL;

	isl_union_set_free(stencil->delta_union_set_to_scan);

	if(stencil->deltaDimInfo){
		int i;
		for(i=0;i<stencil->loopDim;i++){
			if(stencil->deltaDimInfo[i].positive_vector_set)
				free(stencil->deltaDimInfo[i].positive_vector_set);
			if(stencil->deltaDimInfo[i].negative_vector_set)
				free(stencil->deltaDimInfo[i].negative_vector_set);
		}

		free(stencil->deltaDimInfo);
	}

	isl_union_set_free(stencil->deltaInfoSet);

	if(stencil->spaceInfo)
	{
		int i;
		for(i=0; i<stencil->spaceInfoDim;i++){
			if(stencil->spaceInfo[i].vector){
				free(stencil->spaceInfo[i].vector);
			}
			if(stencil->spaceInfo[i].spaceName){
				free(stencil->spaceInfo[i].spaceName);
			}
		}
		free(stencil->spaceInfo);
	}

	isl_set_free(stencil->spaceInfoHyperplanes);

	if(stencil->domainParams){
		free(stencil->domainParams);
	}
	if(stencil->domainParamsBounds){
		free(stencil->domainParamsBounds);
	}

	isl_set_free(stencil->spaceInfoSet);

	isl_union_set_free(stencil->tile);
	isl_union_set_free(stencil->tileOrder);
	isl_union_map_free(stencil->tileRelations);
	isl_union_set_free(stencil->tileRelationsDelta);
	isl_union_set_free(stencil->tileRelationsDeltaToScan);
	isl_union_map_free(stencil->tileScheduleMap);

	isl_union_map_free(stencil->stencilScheduleMap);
	isl_union_map_free(stencil->stencilGlobalSchedule);

	if(stencil->hyperplaneCheckInfo){
		int i;
		for(i=0; i<stencil->spaceInfoDim;i++){
			free(stencil->hyperplaneCheckInfo->statementList[i]);
		}
		free(stencil->hyperplaneCheckInfo->statementList);
		free(stencil->hyperplaneCheckInfo);
	}

	isl_set_free(stencil->stencilScheduleCardSet);
	isl_union_set_free(stencil->stencilScheduleCardUnionSet);

	isl_val_free(stencil->wavefrontSetCardValue);

	free(stencil);

	return NULL;
}

__isl_give stencil_scop* stencil_scop_alloc(isl_ctx *ctx){
	stencil_scop* stencil;

	stencil = isl_alloc_type(ctx, stencil_scop);

	stencil->ctx = ctx;

	stencil->loopDim = 0;
	stencil->loopDimName = 0;
	stencil->domainParams = 0;
	stencil->domainParamsBounds = 0;

	stencil->delta_union_set_to_scan = 0;
	stencil->deltaDimInfo = 0;
	stencil->deltaInfoSet = 0;

	stencil->isNegParam = isl_bool_false;

	stencil->spaceInfoDim = 0;
	stencil->spaceInfo = 0;
	stencil->spaceInfoEquations[0]='\0';
	stencil->spaceInfoEquationsLexico[0]='\0';
	stencil->spaceInfoHyperplanes = 0;
	stencil->spaceInfoSet = 0;

	stencil->tile = 0;
	stencil->tileOrder = 0;
	stencil->tileRelations = 0;
	stencil->tileRelationsDelta = 0;
	stencil->tileRelationsDeltaToScan = 0;
	stencil->tileWfSpaceEquations[0] = '\0';
	stencil->tileWfSpaceExpr[0] = '\0';
	stencil->tileScheduleMap = 0;

	stencil->stencilScheduleMap = 0;
	stencil->stencilGlobalSchedule = 0;

	stencil->hyperplaneCheckInfo = 0;

	stencil->stencilScheduleCardSet = 0;
	stencil->stencilScheduleCardUnionSet = 0;

	stencil->wavefrontSetCardValue = 0;

	return stencil;
}

char* stencil_scop_deltaDim_Info_to_str(__isl_keep stencil_scop *stencil, int pos){
	char buff[1024];
	char* vectorInfoStr = 0;
	buff[0] = '\0';

	if(stencil->deltaDimInfo){
		sprintf(buff,"[%s]: \n", stencil->loopDimName[pos]);

		if(stencil->deltaDimInfo[pos].positive_vector_set)
			sprintf(&buff[strlen(buff)],"positive: %s \n", stencil->deltaDimInfo[pos].positive_vector_set);
		if(stencil->deltaDimInfo[pos].negative_vector_set)
			sprintf(&buff[strlen(buff)],"negative: %s \n", stencil->deltaDimInfo[pos].negative_vector_set);

		if(stencil->deltaDimInfo[pos].isNoUpperBound == isl_bool_true || stencil->deltaDimInfo[pos].isNoLowerBound == isl_bool_true){
			sprintf(&buff[strlen(buff)],"is parameter");
		}
		else{
			sprintf(&buff[strlen(buff)],"is not parameter");
		}
	}

	vectorInfoStr = malloc(sizeof(char)*(strlen(buff)+1));
	sprintf(vectorInfoStr,"%s",buff);

	return vectorInfoStr;
}

char* stencil_scop_space_Info_to_str(__isl_keep stencil_scop *stencil, int pos){
	int i;
	char buff[1024];
	char* spaceInfoStr = 0;
	buff[0] = '\0';

	if(stencil->spaceInfo){
		sprintf(buff,"[space%i]:", pos + 1);

		sprintf(&buff[strlen(buff)],"\nhyperplane:");
		sprintf(&buff[strlen(buff)]," %s * ( %i )", stencil->loopDimName[0],stencil->spaceInfo[pos].vector[0]);
		for(i=1; i<stencil->loopDim; i++){
			sprintf(&buff[strlen(buff)]," + %s * ( %i )", stencil->loopDimName[i],stencil->spaceInfo[pos].vector[i]);
		}
		sprintf(&buff[strlen(buff)]," + b = 0");
	}

	spaceInfoStr = malloc(sizeof(char)*(strlen(buff)+1));
	sprintf(spaceInfoStr,"%s",buff);

	return spaceInfoStr;
}

static void init_info_in_stencil_scop(__isl_keep calc_scop* calc, stencil_scop* stencil){
	{

		char* domainstr = isl_union_set_to_str(calc->domain);
		stencil->domainParams = str_substring(domainstr, "[", "] -> {");

		if(strlen(stencil->domainParams)>0){
			isl_printer *printer = isl_printer_to_str(stencil->ctx);
			char strBuf[256];

			{
				char *tmpBuf;
				sprintf(strBuf, " <= %i and ", stencil->stencilParam->bound);
				tmpBuf = str_substring(domainstr, "[", " -> {");
				tmpBuf = str_replace(tmpBuf,",", strBuf);
				tmpBuf = str_replace(tmpBuf,"]", strBuf);

				printer = isl_printer_print_str(printer, tmpBuf);
			}
			{
				char *tmpBuf;
				sprintf(strBuf, " >= 0 and ");
				tmpBuf = str_substring(domainstr, "[", " -> {");
				tmpBuf = str_replace(tmpBuf,",", strBuf);
				tmpBuf = str_replace(tmpBuf,"]", strBuf);

				printer = isl_printer_print_str(printer, tmpBuf);
			}

			stencil->domainParamsBounds = isl_printer_get_str(printer);
			isl_printer_free(printer);
		}
		else{
			stencil->domainParamsBounds = str_substring(domainstr, "[", "] -> {");
		}

		free(domainstr);
	}

	stencil->loopDim = calc->dimCount;
	stencil->loopDimName = calc->idNames;

	{
		int i;

		stencil->deltaDimInfo = malloc(sizeof(deltaDim_Info)*stencil->loopDim);
		for(i=0; i < stencil->loopDim; i++){
			stencil->deltaDimInfo[i].isNoUpperBound = isl_bool_false;
			stencil->deltaDimInfo[i].isNoLowerBound = isl_bool_false;
			stencil->deltaDimInfo[i].positive_val = 0;
			stencil->deltaDimInfo[i].positive_vector_set = 0;
			stencil->deltaDimInfo[i].negative_val = 0;
			stencil->deltaDimInfo[i].negative_vector_set = 0;
		}
	}


}

static void delta_union_set_analyze(__isl_keep isl_union_set* delta_union_set, stencil_scop* stencil){
	//equations described hyperplane
	{
		int i;

		sprintf(stencil->spaceInfoEquations,"{[d0, b0, c0");
		for(i=0;i<stencil->loopDim;i++){
			sprintf(&(stencil->spaceInfoEquations)[strlen(stencil->spaceInfoEquations)], ", a%i", i + 1);
		}

		sprintf(&(stencil->spaceInfoEquations)[strlen(stencil->spaceInfoEquations)], "] : exists b1");
		sprintf(&(stencil->spaceInfoEquations)[strlen(stencil->spaceInfoEquations)], ", c1");
		for(i=1;i<stencil->loopDim;i++){
			sprintf(&(stencil->spaceInfoEquations)[strlen(stencil->spaceInfoEquations)], ", b%i", i + 1);
			sprintf(&(stencil->spaceInfoEquations)[strlen(stencil->spaceInfoEquations)], ", c%i", i + 1);
		}

		sprintf(&(stencil->spaceInfoEquations)[strlen(stencil->spaceInfoEquations)], " : ((b1 = a1 and a1 >= 0) or (b1 = -a1 and a1 < 0))");
		sprintf(&(stencil->spaceInfoEquations)[strlen(stencil->spaceInfoEquations)], " and ((c1 = -1 and a1 >= 0) or (c1 = 0 and a1 < 0))");
		for(i=1;i<stencil->loopDim;i++){
			sprintf(&(stencil->spaceInfoEquations)[strlen(stencil->spaceInfoEquations)], " and ((b%i = a%i and a%i >= 0) or (b%i = -a%i and a%i < 0))", i + 1, i + 1, i + 1, i + 1, i + 1, i + 1);
			sprintf(&(stencil->spaceInfoEquations)[strlen(stencil->spaceInfoEquations)], " and ((c%i = -1 and a%i >= 0) or (c%i = 0 and a%i < 0))", i + 1, i + 1, i + 1, i + 1);
		}

		sprintf(&(stencil->spaceInfoEquations)[strlen(stencil->spaceInfoEquations)], " and b0 > 0 and b0 = b1");
		for(i=1;i<stencil->loopDim;i++){
			sprintf(&(stencil->spaceInfoEquations)[strlen(stencil->spaceInfoEquations)], " + b%i", i + 1);
		}

		sprintf(&(stencil->spaceInfoEquations)[strlen(stencil->spaceInfoEquations)], " and c0 = c1");
		for(i=1;i<stencil->loopDim;i++){
			sprintf(&(stencil->spaceInfoEquations)[strlen(stencil->spaceInfoEquations)], " + c%i", i + 1);
		}
	}

	isl_debug_printf("\n%s\n", "#delta analyze start");
	//delta analyze
	if(isl_union_set_is_empty(delta_union_set)!=isl_bool_true){
		{
			isl_union_set* delta_union_set_prepared = 0;

			//clear tupleName
			//isl_union_set_foreach_set_in_set(delta_union_set, delta_set_prepare, &delta_union_set_prepared);
			delta_union_set_prepared = isl_union_set_copy(delta_union_set);
			delta_union_set_prepared = isl_union_set_coalesce(delta_union_set_prepared);

			//clear input parameters
			stencil->deltaInfoSet = isl_union_set_project_out_all_params(delta_union_set_prepared);
		}
		//call analyze for all points and scan points and form vectors for spaces
		{
			dimInfo info;

			info.dimPos = 0;
			info.stencil = stencil;
			isl_union_set_foreach_set_in_set(stencil->deltaInfoSet, delta_set_analyze, &info);
		}
		{
			int i;

			for(i=0; i < stencil->loopDim; i++){
				if(stencil->deltaDimInfo[i].isNoLowerBound == isl_bool_true){
					sprintf(&(stencil->spaceInfoEquations)[strlen(stencil->spaceInfoEquations)], " and a%i <= 0 ", i+ 1 );
				}
				if(stencil->deltaDimInfo[i].isNoUpperBound == isl_bool_true){
					sprintf(&(stencil->spaceInfoEquations)[strlen(stencil->spaceInfoEquations)], " and a%i >= 0 ", i+ 1 );
				}
			}

			isl_union_set_foreach_point(stencil->delta_union_set_to_scan, delta_set_point_analyze, stencil);
		}
	}
	else{
		stencil->deltaInfoSet = isl_union_set_read_from_str(stencil->ctx, "{}");
	}

	{
		int i;
		sprintf(stencil->spaceInfoEquationsLexico,"%s", stencil->spaceInfoEquations);

		sprintf(&(stencil->spaceInfoEquationsLexico)[strlen(stencil->spaceInfoEquationsLexico)], " and b0 > 0 and (a1 > 0");
		for(i=1;i<stencil->loopDim;i++){
			int j;
			sprintf(&(stencil->spaceInfoEquationsLexico)[strlen(stencil->spaceInfoEquationsLexico)], " or (a%i > 0", i + 1);
			for(j=i;j>0;j--){
				sprintf(&(stencil->spaceInfoEquationsLexico)[strlen(stencil->spaceInfoEquationsLexico)], " and a%i = 0", j);
			}
			sprintf(&(stencil->spaceInfoEquationsLexico)[strlen(stencil->spaceInfoEquationsLexico)], ")");
		}
		sprintf(&(stencil->spaceInfoEquationsLexico)[strlen(stencil->spaceInfoEquationsLexico)]," ) }");
	}


	if(isl_debug_is_on())
	{
		int i;
		isl_debug_printf("\n%s\n","#analyze loop dimensions:");
		for(i=0; i<stencil->loopDim; i++){
			isl_debug_printf_str("\n#%s\n",str_replace(stencil_scop_deltaDim_Info_to_str(stencil,i), "\n"," "));
		}
	}

	isl_debug_printf("\n%s\n", "#delta analyze end");

	//close equations
	sprintf(&(stencil->spaceInfoEquations)[strlen(stencil->spaceInfoEquations)], " }");

}
static isl_stat delta_set_analyze(__isl_take isl_set *delta_set, void *user){
	dimInfo *info = (dimInfo*) user;
	stencil_scop* stencil = info->stencil;
	int dimPos = info->dimPos;

	if(dimPos < stencil->loopDim){
		char dimNamesList[256];
		isl_bool lower_bound, upper_bound;
		const char* dimName = isl_set_get_dim_name(delta_set, isl_dim_set, dimPos);

		delta_set = isl_set_remove_divs(delta_set);

		dimNamesList[0]='\0';
		{
			int i;
			for(i=0; i < stencil->loopDim; i++ ){
				sprintf(&dimNamesList[strlen(dimNamesList)], ",%s", isl_set_get_dim_name(delta_set,isl_dim_set, i));
			}
		}
		dimNamesList[0] = ' ';

		{
			isl_set* tmpSet = isl_set_copy(delta_set);
			if(dimPos > 0)
			{
				tmpSet = isl_set_remove_dims(tmpSet, isl_dim_set, 0, dimPos);
			}
			if(stencil->loopDim - (dimPos + 1) > 0)
			{
				tmpSet = isl_set_remove_dims(tmpSet, isl_dim_set, 1, stencil->loopDim - (dimPos + 1));
			}

			lower_bound = isl_set_dim_has_lower_bound(tmpSet, isl_dim_set, 0);
			upper_bound = isl_set_dim_has_upper_bound(tmpSet, isl_dim_set, 0);

			isl_set_free(tmpSet);
		}

		{
			char* minValStr = 0;
			char* maxValStr = 0;

			if(lower_bound == isl_bool_false && upper_bound == isl_bool_false){
				stencil->deltaDimInfo[dimPos].isNoLowerBound = isl_bool_true;
				stencil->deltaDimInfo[dimPos].isNoUpperBound = isl_bool_true;
				minValStr = malloc(2*sizeof(char));
				maxValStr = malloc(2*sizeof(char));

				sprintf(minValStr,"0");
				sprintf(maxValStr,"0");
			}
			else if(lower_bound == isl_bool_false){
				char buff[1024];
				maxValStr = get_pw_aff_value_to_str(isl_set_dim_max(isl_set_copy(delta_set), dimPos));
				stencil->deltaDimInfo[dimPos].isNoLowerBound = isl_bool_true;

				sprintf(buff,"{ [%s] : %s < 0 and %s < %s }", dimNamesList, dimName, dimName, maxValStr);

				minValStr = get_pw_aff_value_to_str(isl_set_dim_max(isl_set_intersect(isl_set_copy(delta_set), isl_set_read_from_str(stencil->ctx, buff)), dimPos));
			}
			else if(upper_bound == isl_bool_false){
				char buff[1024];
				minValStr = get_pw_aff_value_to_str(isl_set_dim_min(isl_set_copy(delta_set), dimPos));
				stencil->deltaDimInfo[dimPos].isNoUpperBound = isl_bool_true;

				sprintf(buff,"{ [%s] : 0 < %s and %s < %s }", dimNamesList, dimName, minValStr, dimName);

				maxValStr = get_pw_aff_value_to_str(isl_set_dim_min(isl_set_intersect(isl_set_copy(delta_set), isl_set_read_from_str(stencil->ctx, buff)), dimPos));
			}
			else{
				minValStr = get_pw_aff_value_to_str(isl_set_dim_min(isl_set_copy(delta_set), dimPos));
				maxValStr = get_pw_aff_value_to_str(isl_set_dim_max(isl_set_copy(delta_set), dimPos));
			}

			{
				char buff[1024];

				sprintf(buff,"{ [%s] : %s = %s or 0 = %s or %s = %s }", dimNamesList , minValStr, dimName, dimName, maxValStr, dimName );
				delta_set = isl_set_intersect(delta_set, isl_set_read_from_str(stencil->ctx, buff));
			}

			free(minValStr);
			free(maxValStr);
		}

		{
			dimInfo info;
			info.dimPos = dimPos + 1;
			info.stencil = stencil;

			isl_union_set *unionSet = isl_union_set_from_set(delta_set);
			isl_union_set_foreach_set_in_set(unionSet, delta_set_analyze, &info);
			isl_union_set_free(unionSet);
		}
	}
	else{
		isl_debug_printf_set("\n#distance vectors set : %s\n", delta_set);

		if(stencil->delta_union_set_to_scan)
			stencil->delta_union_set_to_scan = isl_union_set_union(stencil->delta_union_set_to_scan, isl_union_set_from_set(delta_set));
		else
			stencil->delta_union_set_to_scan = isl_union_set_from_set(delta_set);
	}

	return isl_stat_ok;
}

static isl_stat delta_set_point_analyze(__isl_take isl_point *pnt, void *user){
	int i;
	stencil_scop* stencil = (stencil_scop*) user;
	isl_set* set = isl_set_from_point(pnt);

	int dim0Value = get_pw_aff_value_to_int( isl_set_dim_max(isl_set_copy(set), 0 ));

	sprintf(&(stencil->spaceInfoEquations)[strlen(stencil->spaceInfoEquations)], " and %i*a1", dim0Value);

	for(i= 1; i < stencil->loopDim; i++){
		int dimIValue = get_pw_aff_value_to_int( isl_set_dim_max(isl_set_copy(set), i));

		if(dimIValue>0){
			float value = (float)dim0Value / (float)dimIValue;
			if(stencil->deltaDimInfo[i].positive_vector_set){
				if(stencil->deltaDimInfo[i].positive_val > value){
					stencil->deltaDimInfo[i].positive_val = value;
					free(stencil->deltaDimInfo[i].positive_vector_set);
					stencil->deltaDimInfo[i].positive_vector_set = isl_set_to_str(set);
				}
			} else{
				stencil->deltaDimInfo[i].positive_val = value;
				stencil->deltaDimInfo[i].positive_vector_set = isl_set_to_str(set);
			}
		}

		if(dimIValue<0){
			float value = (float)dim0Value / (float)(-1*dimIValue);
			if(stencil->deltaDimInfo[i].negative_vector_set){
				if(stencil->deltaDimInfo[i].negative_val > value){
					stencil->deltaDimInfo[i].negative_val = value;
					free(stencil->deltaDimInfo[i].negative_vector_set);
					stencil->deltaDimInfo[i].negative_vector_set = isl_set_to_str(set);
				}
			} else{
				stencil->deltaDimInfo[i].negative_val = value;
				stencil->deltaDimInfo[i].negative_vector_set = isl_set_to_str(set);
			}

		}

		sprintf(&(stencil->spaceInfoEquations)[strlen(stencil->spaceInfoEquations)], " + %i*a%i", dimIValue, i+1);
	}

	sprintf(&(stencil->spaceInfoEquations)[strlen(stencil->spaceInfoEquations)], " >= 0");

	isl_debug_printf_set("\n#selected distance vector: %s\n", set);

	isl_set_free(set);

	return isl_stat_ok;
}

static void space_info_create_tile(__isl_keep isl_union_set *domain_union_set, stencil_scop* stencil){
	isl_debug_printf("\n%s\n", "#create space start");
	//create spaceInfoHyperplanes
	if(stencil->stencilParam->method == 1)
	{
		int i;
		for(i=0;i<stencil->loopDim;i++){
			space_info_create_hyperplane_for_dim(stencil, i);
		}
	}
	if(stencil->stencilParam->method == 2)
	{
		int i;
		if(stencil->deltaDimInfo)
		{
			isl_union_set* vectors = isl_union_set_read_from_str(stencil->ctx,"{}");

			for(i=0; i<stencil->loopDim;i++){
				if(stencil->deltaDimInfo[i].positive_vector_set && stencil->deltaDimInfo[i].negative_vector_set){
					vectors = isl_union_set_union(vectors, isl_union_set_read_from_str(stencil->ctx,stencil->deltaDimInfo[i].positive_vector_set));
					vectors = isl_union_set_union(vectors, isl_union_set_read_from_str(stencil->ctx,stencil->deltaDimInfo[i].negative_vector_set));
				}
			}

			isl_union_set_foreach_point(vectors, space_info_create_hyperplane_for_delta_vector, stencil);
			isl_union_set_free(vectors);
		}
		stencil->spaceInfoHyperplanesDims = malloc(sizeof(int)*stencil->loopDim);

		for(i=0; i<stencil->loopDim;i++){
				stencil->spaceInfoHyperplanesDims[i] = 0;
		}

		isl_set_foreach_point(stencil->spaceInfoHyperplanes, space_info_hyperplanes_dims_check, stencil);

		for(i=0;i<stencil->loopDim;i++){
			if(!stencil->spaceInfoHyperplanesDims[i]){
				space_info_create_hyperplane_for_dim(stencil, i);
			}
		}
	}
	if(stencil->stencilParam->method == 3){
		space_info_create_hyperplane_max_non_zero_dim(stencil);
		space_info_create_hyperplane_for_dim(stencil, 0);
	}

	//memory allocating
	{
		int i;

		stencil->spaceInfoDim = 0;
		isl_set_foreach_point(stencil->spaceInfoHyperplanes, space_info_hyperplanes_count, stencil);

		stencil->spaceInfo = malloc(sizeof(space_Info)*stencil->spaceInfoDim);
		for(i=0; i<stencil->spaceInfoDim;i++){
			stencil->spaceInfo[i].vector = malloc(sizeof(int)*stencil->loopDim);
			stencil->spaceInfo[i].spaceName = 0;
		}

		stencil->spaceInfoDim = 0;

		isl_set_foreach_point(stencil->spaceInfoHyperplanes, space_info_from_hyperplanes, stencil);
	}
	//set space index name
	{
		int i,idx = 1;
		char* domainStr = isl_union_set_to_str(domain_union_set);
		char bufStr[128];

		for(i=0;i<stencil->spaceInfoDim;i++){
			sprintf(bufStr,"Id%i", idx);
			while(strstr(domainStr,bufStr)){
				idx += 1;
				sprintf(bufStr,"Id%i", idx);
			}
			stencil->spaceInfo[i].spaceName = malloc(sizeof(char)*(strlen(bufStr)+1));
			sprintf(stencil->spaceInfo[i].spaceName,"%s",bufStr);
			idx += 1;
		}

		free(domainStr);
	}

	{
		int i;
		for(i=0; i<stencil->spaceInfoDim; i++){
			space_info_add_space_to_set(stencil,i);
		}
	}

	if(stencil->spaceInfoDim > 0){
		isl_union_set_foreach_set_in_set(domain_union_set, space_info_build_tile_domain, stencil);
	}

#if defined PRINT_DEBUG_INFO
	{
		isl_debug_printf("\n%s\n", "#created spaces:");
		int i;
		for(i=0; i<stencil->spaceInfoDim; i++){
			isl_debug_printf_str("\n#%s\n",str_replace(stencil_scop_space_Info_to_str(stencil,i),"\n"," "));
		}
	}
#endif

	isl_debug_printf("\n%s\n", "#create space end");
}

static void space_info_create_hyperplane_for_dim(stencil_scop* stencil, int dim){
	int i;
	char equationStr[4096];

	sprintf(equationStr,"{[d0 = 0, b0, c0");
	for(i=0;i<stencil->loopDim;i++){
		sprintf(&equationStr[strlen(equationStr)], ", a%i", i + 1);
	}

	sprintf(&equationStr[strlen(equationStr)], "] : a%i != 0 }", dim + 1);

	{
		isl_set* ortogonalVector = isl_set_lexmin(isl_set_intersect(isl_set_read_from_str(stencil->ctx, stencil->spaceInfoEquations), isl_set_read_from_str(stencil->ctx, equationStr)));

		isl_debug_printf("\n#hyperplane_for_dim: lexmin( %s * ", equationStr);
		isl_debug_printf("%s )\n", stencil->spaceInfoEquations);

		isl_debug_printf_set("\n#result: %s\n", ortogonalVector);

		if(space_info_check_hyperplane(stencil, ortogonalVector) == isl_bool_true){
			if(!stencil->spaceInfoHyperplanes){
				stencil->spaceInfoHyperplanes = isl_set_copy(ortogonalVector);
			}

			stencil->spaceInfoHyperplanes = isl_set_coalesce(isl_set_union(stencil->spaceInfoHyperplanes, ortogonalVector));
		}
		else
			isl_set_free(ortogonalVector);
	}
}

static isl_bool space_info_check_hyperplane(stencil_scop* stencil, __isl_keep isl_set *hyperplane){
	isl_bool result = isl_bool_false;

	if(isl_set_is_empty(hyperplane) == isl_bool_true){
		isl_debug_printf("\n#hyperplane check: %s \n", "hyperplane is empty");
	}
	else{
		if(!stencil->hyperplaneCheckInfo){
			space_info_add_hyperplane_to_check_info(stencil, hyperplane);
			result = isl_bool_true;

			isl_debug_printf("\n#hyperplane check: %s \n", "first hyperplane");
		}
		else{
			int i;
			char equationStr[4096];

			equationStr[0]='\0';

			sprintf(&equationStr[strlen(equationStr)], "{ : exists %s, c%i ", stencil->hyperplaneCheckInfo->paramIdList, stencil->hyperplaneCheckInfo->lastId + 1);
			sprintf(&equationStr[strlen(equationStr)], ": ( %s or c%i!=0 ) ", stencil->hyperplaneCheckInfo->paramIdCondition, stencil->hyperplaneCheckInfo->lastId + 1);
			for(i=0;i<stencil->loopDim;i++){
				sprintf(&equationStr[strlen(equationStr)], "and c%i*%i = %s ", stencil->hyperplaneCheckInfo->lastId + 1, get_pw_aff_value_to_int( isl_set_dim_max(isl_set_copy(hyperplane), i + hyperplaneFrirstIndex)), stencil->hyperplaneCheckInfo->statementList[i]);
			}
			sprintf(&equationStr[strlen(equationStr)], "}");

			isl_set* testSet = isl_set_read_from_str(stencil->ctx, equationStr);

			if(isl_set_is_empty(testSet) == isl_bool_true){
				space_info_add_hyperplane_to_check_info(stencil, hyperplane);
				result = isl_bool_true;
			}

			isl_debug_printf("\n#hyperplane check: %s = {}\n", equationStr);

			isl_set_free(testSet);
		}
	}

	if( result == isl_bool_true){
		isl_debug_printf("\n#result: %s\n", "true");
	}
	else{
		isl_debug_printf("\n#result: %s\n", "false");
	}

	return result;
}

static isl_stat space_info_create_hyperplane_for_delta_vector(__isl_take isl_point *pnt, void *user){
	int i;
	char equationStr[4096];
	stencil_scop* stencil = (stencil_scop*) user;
	isl_set* set = isl_set_from_point(pnt);

	{
		sprintf(equationStr,"{[d0 = 0, b0, c0");
		for(i=0;i<stencil->loopDim;i++){
			sprintf(&equationStr[strlen(equationStr)], ", a%i", i + 1);
		}

		sprintf(&equationStr[strlen(equationStr)], "] : %i * a1", get_pw_aff_value_to_int( isl_set_dim_max(isl_set_copy(set), 0)));
			for(i=1; i<stencil->loopDim; i++){
				sprintf(&equationStr[strlen(equationStr)], " + %i * a%i", get_pw_aff_value_to_int( isl_set_dim_max(isl_set_copy(set), i)), i + 1);
			}
			sprintf(&equationStr[strlen(equationStr)], " = 0 }");

		space_info_hyperplane_build_equation(0, equationStr, 1, stencil);
	}

	{
		isl_set* ortogonalVector = isl_set_lexmin(isl_set_intersect(isl_set_read_from_str(stencil->ctx, stencil->spaceInfoEquations), isl_set_read_from_str(stencil->ctx, equationStr)));

		isl_debug_printf_set("\n#hyperplane_for_distance_vector: %s :", set);
		isl_debug_printf(" lexmin( %s * ", equationStr);
		isl_debug_printf("%s )\n", stencil->spaceInfoEquations);
		isl_debug_printf_set("\n#result: %s\n", ortogonalVector);

		if(space_info_check_hyperplane(stencil, ortogonalVector) == isl_bool_true){
			if(!stencil->spaceInfoHyperplanes){
				stencil->spaceInfoHyperplanes = isl_set_copy(ortogonalVector);
			}

			stencil->spaceInfoHyperplanes = isl_set_coalesce(isl_set_union(stencil->spaceInfoHyperplanes, ortogonalVector));
		}
		else
			isl_set_free(ortogonalVector);
	}
	isl_set_free(set);

	return isl_stat_ok;
}

static void space_info_create_hyperplane_max_non_zero_dim(stencil_scop* stencil){
	int i;
	char equationStr[4096];

	sprintf(equationStr,"{[d0, b0, c0");
	for(i=0;i<stencil->loopDim;i++){
		sprintf(&equationStr[strlen(equationStr)], ", a%i", i + 1);
	}

	sprintf(&equationStr[strlen(equationStr)], "] : exists d1");
	for(i=1;i<stencil->loopDim;i++){
		sprintf(&equationStr[strlen(equationStr)], ", d%i", i + 1);
	}

	sprintf(&equationStr[strlen(equationStr)], " : ((d1 = -1 and a1 != 0) or (d1 = 0 and a1 = 0))");
	for(i=1;i<stencil->loopDim;i++){
		sprintf(&equationStr[strlen(equationStr)], " and ((d%i = -1 and a%i != 0) or (d%i = 0 and a%i = 0))", i + 1, i + 1, i + 1, i + 1);
	}

	sprintf(&equationStr[strlen(equationStr)], " and d0 = d1");
	for(i=1;i<stencil->loopDim;i++){
		sprintf(&equationStr[strlen(equationStr)], " + d%i", i + 1);
	}

	sprintf(&equationStr[strlen(equationStr)], "}");

	{
		isl_set* ortogonalVector = isl_set_lexmin(isl_set_intersect(isl_set_read_from_str(stencil->ctx, stencil->spaceInfoEquations), isl_set_read_from_str(stencil->ctx, equationStr)));

		isl_debug_printf("\n#hyperplane_max_non_zero_dim: lexmin( %s * ", equationStr);
		isl_debug_printf("%s )\n", stencil->spaceInfoEquations);

		isl_debug_printf_set("\n#result: %s\n", ortogonalVector);

		if(space_info_check_hyperplane(stencil, ortogonalVector) == isl_bool_true){
			if(!stencil->spaceInfoHyperplanes){
				stencil->spaceInfoHyperplanes = isl_set_copy(ortogonalVector);
			}

			stencil->spaceInfoHyperplanes = isl_set_coalesce(isl_set_union(stencil->spaceInfoHyperplanes, ortogonalVector));
		}
		else
			isl_set_free(ortogonalVector);
	}
}

static void space_info_add_hyperplane_to_check_info(stencil_scop* stencil, __isl_keep isl_set *hyperplane){

	int i;
	if(!stencil->hyperplaneCheckInfo){
		stencil->hyperplaneCheckInfo = malloc(sizeof(hyperplane_check_info));
		stencil->hyperplaneCheckInfo->lastId = 0;
		stencil->hyperplaneCheckInfo->paramIdList[0] = '\0';
		stencil->hyperplaneCheckInfo->paramIdCondition[0] = '\0';
		stencil->hyperplaneCheckInfo->statementList = (char**)malloc(sizeof(char*)*stencil->loopDim);
		for(i=0; i<stencil->loopDim; i++){
			stencil->hyperplaneCheckInfo->statementList[i] = (char*)malloc(sizeof(char)*256);
			stencil->hyperplaneCheckInfo->statementList[i][0] = '\0';
		}
	}
	stencil->hyperplaneCheckInfo->lastId += 1;

	if(stencil->hyperplaneCheckInfo->lastId==1){
		sprintf(&stencil->hyperplaneCheckInfo->paramIdList[strlen(stencil->hyperplaneCheckInfo->paramIdList)],"c%i", stencil->hyperplaneCheckInfo->lastId);
		sprintf(&stencil->hyperplaneCheckInfo->paramIdCondition[strlen(stencil->hyperplaneCheckInfo->paramIdCondition)],"c%i!=0", stencil->hyperplaneCheckInfo->lastId);
		for(i=0; i<stencil->loopDim; i++){
			sprintf(&stencil->hyperplaneCheckInfo->statementList[i][strlen(stencil->hyperplaneCheckInfo->statementList[i])],"c%i*%i", stencil->hyperplaneCheckInfo->lastId, get_pw_aff_value_to_int( isl_set_dim_max(isl_set_copy(hyperplane), i + hyperplaneFrirstIndex)));
		}
	}
	else{
		sprintf(&stencil->hyperplaneCheckInfo->paramIdList[strlen(stencil->hyperplaneCheckInfo->paramIdList)],", c%i", stencil->hyperplaneCheckInfo->lastId);
		sprintf(&stencil->hyperplaneCheckInfo->paramIdCondition[strlen(stencil->hyperplaneCheckInfo->paramIdCondition)]," or c%i!=0", stencil->hyperplaneCheckInfo->lastId);
		for(i=0; i<stencil->loopDim; i++){
			sprintf(&stencil->hyperplaneCheckInfo->statementList[i][strlen(stencil->hyperplaneCheckInfo->statementList[i])]," + c%i*%i", stencil->hyperplaneCheckInfo->lastId, get_pw_aff_value_to_int( isl_set_dim_max(isl_set_copy(hyperplane), i + hyperplaneFrirstIndex)));
		}
	}
}

static void space_info_hyperplane_build_equation(__isl_take isl_set* set, char* equationStr, int level, stencil_scop* stencil){
	int i;

	if(set){
		equationStr[strlen(equationStr)-1] = '\0';

		sprintf(&equationStr[strlen(equationStr)], "and %i * a1", get_pw_aff_value_to_int( isl_set_dim_max(isl_set_copy(set), hyperplaneFrirstIndex)));
		for(i=1; i< stencil->loopDim; i++){
			sprintf(&equationStr[strlen(equationStr)], " + %i * a%i", get_pw_aff_value_to_int( isl_set_dim_max(isl_set_copy(set), i + hyperplaneFrirstIndex)), i+1);
		}
		sprintf(&equationStr[strlen(equationStr)], " = 0 }");

		isl_set_free(set);
	}

	if(stencil->loopDim > level + 1)
	{
		isl_set* ortogonalVector = isl_set_lexmin(isl_set_intersect(isl_set_read_from_str(stencil->ctx, stencil->spaceInfoEquationsLexico), isl_set_read_from_str(stencil->ctx, equationStr)));
		space_info_hyperplane_build_equation(ortogonalVector, equationStr, level + 1, stencil );
	}
}

static isl_stat space_info_hyperplanes_dims_check(__isl_take isl_point *pnt, void *user){
	int i, value;
	stencil_scop* stencil = (stencil_scop*) user;
	isl_set* set = isl_set_from_point(pnt);

	for(i=0;i<stencil->loopDim;i++){
		if(!stencil->spaceInfoHyperplanesDims[i]){
			value = get_pw_aff_value_to_int( isl_set_dim_max(isl_set_copy(set), i + hyperplaneFrirstIndex));
			if(value != 0){
				stencil->spaceInfoHyperplanesDims[i] = 1;
			}
		}
	}

	isl_set_free(set);

	return isl_stat_ok;
}

static isl_stat space_info_hyperplanes_count(__isl_take isl_point *pnt, void *user){
	stencil_scop* stencil = (stencil_scop*) user;

	stencil->spaceInfoDim += 1;

	isl_point_free(pnt);

	return isl_stat_ok;
}


static isl_stat space_info_from_hyperplanes(__isl_take isl_point *pnt, void *user){
	stencil_scop* stencil = (stencil_scop*) user;
	isl_set* set = isl_set_from_point(pnt);
	int i, currentSpace = stencil->spaceInfoDim;

	for(i=0; i<stencil->loopDim; i++){
		stencil->spaceInfo[currentSpace].vector[i] = get_pw_aff_value_to_int( isl_set_dim_max(isl_set_copy(set), i + hyperplaneFrirstIndex));
	}

	stencil->spaceInfoDim = currentSpace + 1;

	isl_debug_printf_set("\n#created hyperplane : %s\n", set);

	isl_set_free(set);

	return isl_stat_ok;
}

static void space_info_add_space_to_set(stencil_scop* stencil, int spaceInfoPos){
	int i;
	char equationStr[4096];
	space_Info* spaceInfo = &(stencil->spaceInfo[spaceInfoPos]);
	int* spaceInfoVector = spaceInfo->vector;
	int spaceSize = stencil->stencilParam->size;

	if(stencil->stencilParam->sizesDimIndex < stencil->stencilParam->sizesDim){
		spaceSize = stencil->stencilParam->sizes[stencil->stencilParam->sizesDimIndex];
		stencil->stencilParam->sizesDimIndex += 1;
	}

	sprintf(equationStr,"[");
	if(strlen(stencil->domainParams)>0){
		sprintf(&equationStr[strlen(equationStr)],"%s, ", stencil->domainParams);
	}
	sprintf(&equationStr[strlen(equationStr)],"%s] -> { [%s", spaceInfo->spaceName, stencil->loopDimName[0]);
	for(i=1; i<stencil->loopDim; i++){
		sprintf(&equationStr[strlen(equationStr)],", %s", stencil->loopDimName[i]);
	}
	sprintf(&equationStr[strlen(equationStr)],"] : exists b%s : %i * %s", spaceInfo->spaceName, spaceInfoVector[0], stencil->loopDimName[0]);
	for(i=1; i<stencil->loopDim; i++){
		sprintf(&equationStr[strlen(equationStr)]," + %i * %s", spaceInfoVector[i], stencil->loopDimName[i]);
	}
	sprintf(&equationStr[strlen(equationStr)]," + b%s = 0 and ", spaceInfo->spaceName);

	sprintf(&equationStr[strlen(equationStr)],"%i * %s", spaceSize, spaceInfo->spaceName);
	sprintf(&equationStr[strlen(equationStr)]," <= b%s <= ", spaceInfo->spaceName);
	sprintf(&equationStr[strlen(equationStr)],"%i * (%s + 1) - 1 }", spaceSize, spaceInfo->spaceName);

	{
		isl_set* space;
		space = isl_set_read_from_str(stencil->ctx, equationStr);

		isl_debug_printf("\n# %s \n", equationStr );

		if(!stencil->spaceInfoSet){
			stencil->spaceInfoSet = isl_set_copy(space);
		}

		stencil->spaceInfoSet = isl_set_coalesce(isl_set_intersect(stencil->spaceInfoSet, space));
	}
}

static isl_stat space_info_build_tile_domain(__isl_take isl_set *domainSet, void *user){
	stencil_scop* stencil = (stencil_scop*)user;

	const char* tupleName = isl_set_get_tuple_name(domainSet);
	isl_set* spacesSet = isl_set_set_tuple_name(isl_set_copy(stencil->spaceInfoSet), tupleName);
	isl_set* tile = isl_set_intersect(isl_set_copy(spacesSet), isl_set_copy(domainSet));

	isl_debug_printf_set("\n#tiles for domain set  : %s\n", domainSet);
	isl_debug_printf_set("\n# %s *", spacesSet);
	isl_debug_printf_set(" %s\n", domainSet);
	isl_debug_printf_set("\n#result  : %s\n", tile);


	{
		if(!stencil->tile){
			stencil->tile = isl_union_set_from_set( isl_set_copy(tile));
		}

		stencil->tile = isl_union_set_coalesce(isl_union_set_union(stencil->tile, isl_union_set_from_set(tile)));
	}

	isl_set_free(domainSet);
	isl_set_free(spacesSet);

	return isl_stat_ok;
}

static void tile_relations_build_map(isl_union_map* relations, stencil_scop* stencil){
	char emptyRelations[128];

	isl_debug_printf("\n%s\n", "#create tile relations start");

	sprintf(emptyRelations,"[%s] -> {}",stencil->domainParams);

	if(isl_union_map_is_empty(relations)== isl_bool_true){
		stencil->tileRelations = isl_union_map_read_from_str(stencil->ctx, emptyRelations);
		isl_debug_printf_union_map("\n#tile relations: %s\n", stencil->tileRelations);
		return;
	}

	{
		int i;
		char substr[128];
		char replacement[256];
		char* tileStr = isl_union_set_to_str(stencil->tile);

		{
			char* strFind = str_substring(tileStr, "[","] ->");
			char* strTo = str_substring(tileStr, "[","] ->");

			for(i=0; i<stencil->spaceInfoDim;i++){
				sprintf(substr,"%s", stencil->spaceInfo[i].spaceName);
				sprintf(replacement,"%s', %s", stencil->spaceInfo[i].spaceName, stencil->spaceInfo[i].spaceName);
				strTo = str_replace(strTo, substr, replacement);
			}

			tileStr = str_replace(tileStr, strFind, strTo);

			free(strFind);
			free(strTo);
		}

		{
			sprintf(replacement," and not (%s=%s'", stencil->spaceInfo[0].spaceName, stencil->spaceInfo[0].spaceName);
			for(i=1; i<stencil->spaceInfoDim;i++){
				sprintf(&replacement[strlen(replacement)]," and %s=%s'", stencil->spaceInfo[i].spaceName, stencil->spaceInfo[i].spaceName);
			}
			sprintf(&replacement[strlen(replacement)]," )");

			sprintf(substr,";");
			sprintf(&replacement[strlen(replacement)],";");
			tileStr = str_replace(tileStr, substr, replacement);

			sprintf(substr,"}");
			sprintf(&replacement[strlen(replacement)-1],"}");
			tileStr = str_replace(tileStr, substr, replacement);
		}


		isl_debug_printf("\n#tlieOrder: %s\n", tileStr);
		stencil->tileOrder = isl_union_set_read_from_str(stencil->ctx, tileStr);

		free(tileStr);
	}
	{
		char strBuf[128];

		sprintf(strBuf,"[%s] -> { }", stencil->domainParams);
		stencil->tileRelations = isl_union_map_read_from_str(stencil->ctx, strBuf);
	}

	isl_union_map_foreach_map_in_map(relations, tile_relations_from_loop_relations, stencil);

	isl_debug_printf_union_map("\n#tile relations: %s\n", stencil->tileRelations);
	isl_debug_printf("\n%s\n", "#create tile relations end");
}

static isl_stat tile_relations_from_loop_relations(__isl_take isl_map *relation, void *user){
	stencil_scop* stencil = (stencil_scop*) user;
	isl_union_map* r_tileTmp = 0;

	isl_debug_printf_map("\n#rlation: %s\n", relation);
	r_tileTmp = isl_union_map_intersect_domain(isl_union_map_from_map(isl_map_copy(relation)),isl_union_set_copy(stencil->tile));
	r_tileTmp = isl_union_map_coalesce(isl_union_map_intersect_range(r_tileTmp, isl_union_set_copy(stencil->tileOrder)));

	isl_debug_printf_union_map("\n#(rlation * tile)->*tlieOrder: %s\n", r_tileTmp);

	isl_union_map_foreach_map_in_map(r_tileTmp, tile_relations_from_map, stencil);


	isl_union_map_free(r_tileTmp);
	isl_map_free(relation);

	return isl_stat_ok;
}

static isl_stat tile_relations_from_map(__isl_take isl_map *map, void *user){
	char* r_tileStr = isl_map_to_str(map);
	char spaceNameListFrom[1024];
	char spaceNameListTo[1024];
	char indexNameList[1024];

	stencil_scop* stencil = (stencil_scop*) user;

	spaceNameListFrom[0] = '\0';
	spaceNameListTo[0] = '\0';
	indexNameList[0] = '\0';

	{
		int i;
		char substr[128];

		for(i=0; i<stencil->spaceInfoDim;i++){
			sprintf(substr,"%s, %s'", stencil->spaceInfo[i].spaceName, stencil->spaceInfo[i].spaceName);
			r_tileStr = str_replace(r_tileStr, substr, "");
			r_tileStr = str_replace(r_tileStr, ", ]", "]");
			sprintf(&spaceNameListFrom[strlen(spaceNameListFrom)],"%s,", stencil->spaceInfo[i].spaceName);
			sprintf(&spaceNameListTo[strlen(spaceNameListTo)],"%s',", stencil->spaceInfo[i].spaceName);
		}
		for(i=0; i<stencil->loopDim;i++){
			sprintf(&indexNameList[strlen(indexNameList)],"%s,%s',", stencil->loopDimName[i], stencil->loopDimName[i]);
		}

		spaceNameListFrom[strlen(spaceNameListFrom)-1] = '\0';
		spaceNameListTo[strlen(spaceNameListTo)-1] = '\0';
		indexNameList[strlen(indexNameList)-1] = '\0';
	}
	{
		char substr[1024];
		char replacement[1024];
		const char* tupleNameIn = isl_map_get_tuple_name(map,isl_dim_in);
		const char* tupleNameOut = isl_map_get_tuple_name(map,isl_dim_out);
		int startPos = 0;
		char* tokStart;
		char* tokEnd;


		sprintf(replacement,"{ %s[%s] -> %s[%s] : exists %s :",tupleNameIn, spaceNameListFrom, tupleNameOut, spaceNameListTo, indexNameList);

		tokStart = strstr(r_tileStr, "{");
		tokEnd = strstr(r_tileStr, ":");
		startPos = tokStart - r_tileStr;

		memcpy(&substr[0], r_tileStr + startPos ,(tokEnd - tokStart)+1 );
		substr[(tokEnd - tokStart) + 1] = '\0';
		r_tileStr = str_replace(r_tileStr,  substr, replacement);
	}

	{
		isl_union_map* tileRelations;
		tileRelations = isl_union_map_read_from_str(isl_map_get_ctx(map), r_tileStr);
		isl_debug_printf("\n#tile relation: %s\n",r_tileStr);

		if(!stencil->tileRelations){
			stencil->tileRelations = isl_union_map_copy(tileRelations);
		}

		stencil->tileRelations = isl_union_map_coalesce(isl_union_map_union(stencil->tileRelations, tileRelations));
	}

	free(r_tileStr);
	isl_map_free(map);

	return isl_stat_ok;
}

static void tile_relations_delta_stencil_schedule_create(__isl_keep isl_union_map* schedule_map, stencil_scop* stencil, int scopeDimSize){
	isl_set* ortogonalVector = 0;

	isl_debug_printf("%s", "\n#create stencil schedule start\n");

	tile_relations_delta_create(stencil);

	isl_debug_printf_union_set("\n#tile relatiions delta: %s\n",stencil->tileRelationsDelta);


	stencil->tileRelationsDeltaToScan = isl_union_set_read_from_str(stencil->ctx, "{ }");
	{
		dimInfo info;
		info.dimPos = 0;
		info.stencil = stencil;
		isl_union_set_foreach_set_in_set(stencil->tileRelationsDelta, tile_relations_delta_set_to_vectors, &info);
	}

	isl_debug_printf_union_set("\n#tile relatiions distance vectors to scan: %s\n", stencil->tileRelationsDeltaToScan);

	ortogonalVector = tile_relations_delta_create_wafefront_hyperplane(stencil);

	{
		int i;
		char bufWfStr[4096];
		char* strScheduleMap = isl_union_map_to_str(schedule_map);

		{
			bufWfStr[0] = '\0';
			sprintf(&bufWfStr[strlen(bufWfStr)],"[%s] -> { [", stencil->domainParams );
			for(i = 0; i < stencil->spaceInfoDim; i++){
				sprintf(&bufWfStr[strlen(bufWfStr)],"%s, ", stencil->spaceInfo[i].spaceName);
			}
			sprintf(&bufWfStr[strlen(bufWfStr)-2],"] -> [");

			for(i = 0; i < stencil->spaceInfoDim; i++){
				char * value = get_pw_aff_value_to_str(isl_set_dim_min(isl_set_copy(ortogonalVector), i + 3));
				sprintf(&bufWfStr[strlen(bufWfStr)],"%s * %s, ", value, stencil->spaceInfo[i].spaceName);
				free(value);
			}
			sprintf(&bufWfStr[strlen(bufWfStr)-2],"] }");

			stencil->tileScheduleMap = isl_union_map_read_from_str(stencil->ctx, bufWfStr);
		}

		bufWfStr[0]='\0';
		sprintf(&bufWfStr[strlen(bufWfStr)],"] -> [");
		{
			int idx = 1;

			sprintf(stencil->wfIndex,"Wid%i", idx);
			while(strstr(strScheduleMap,stencil->wfIndex)){
				idx += 1;
				sprintf(stencil->wfIndex,"Wid%i", idx);
			}
		}

		sprintf(&bufWfStr[strlen(bufWfStr)],"%i, %s, ", stencil->stencil_index, stencil->wfIndex);
		for(i = 0; i < stencil->spaceInfoDim; i++){
			sprintf(&bufWfStr[strlen(bufWfStr)],"%s, ", stencil->spaceInfo[i].spaceName);
		}

		for(i = 0; i < scopeDimSize - stencil->spaceInfoDim; i++){
			sprintf(&bufWfStr[strlen(bufWfStr)],"%s, ", "0");
		}

		strScheduleMap = str_replace(strScheduleMap, "] -> [", bufWfStr);


		sprintf(&stencil->tileWfSpaceExpr[strlen(stencil->tileWfSpaceExpr)],"%s = ", stencil->wfIndex);

		for(i = 0; i < stencil->spaceInfoDim; i++){
			char * value = get_pw_aff_value_to_str(isl_set_dim_min(isl_set_copy(ortogonalVector), i + 3));
			sprintf(&stencil->tileWfSpaceExpr[strlen(stencil->tileWfSpaceExpr)],"%s * %s + ", value, stencil->spaceInfo[i].spaceName);
			free(value);
		}
		sprintf(&stencil->tileWfSpaceExpr[strlen(stencil->tileWfSpaceExpr)],"0 and ");

		bufWfStr[0]='\0';
		sprintf(bufWfStr," : %s", stencil->tileWfSpaceExpr);
		{
			char *spaceInfoSetStr = isl_set_to_str(stencil->spaceInfoSet);

			{
				isl_basic_set* spaceInfoBasicSet = isl_basic_set_read_from_str(stencil->ctx, spaceInfoSetStr);

				free(spaceInfoSetStr);
				spaceInfoSetStr = isl_basic_set_to_str(spaceInfoBasicSet);
				isl_basic_set_free(spaceInfoBasicSet);
			}

			{
				char* strEquation = str_substring(spaceInfoSetStr, ":", "}");

				sprintf(&bufWfStr[strlen(bufWfStr)],"%s ;", strEquation);
				strScheduleMap[strlen(strScheduleMap)-1] = ';';
				strScheduleMap = str_replace(strScheduleMap, ";", bufWfStr);
				strScheduleMap[strlen(strScheduleMap)-1] = '}';
				isl_debug_printf("\n#stencil schedule map: %s \n ", strScheduleMap);
				stencil->stencilScheduleMap = isl_union_map_read_from_str(stencil->ctx, strScheduleMap);

				free(strEquation);
			}

			free(spaceInfoSetStr);
		}

		free(strScheduleMap);
	}

	isl_debug_printf("%s", "\n#create stencil schedule end\n");

	isl_set_free(ortogonalVector);
}

static void tile_relations_delta_create(stencil_scop* stencil){
	isl_union_map* relations = 0;

	{
		char strBuf[128];

		sprintf(strBuf,"[%s] -> { }", stencil->domainParams);
		relations = isl_union_map_read_from_str(stencil->ctx, strBuf);
	}


	isl_union_map_foreach_map_in_map(stencil->tileRelations, clear_tuple_names_in_map, &relations);
	relations = isl_union_map_project_out_all_params(relations);

	stencil->tileRelationsDelta = isl_union_set_coalesce(isl_union_map_deltas(relations));
}

static isl_stat tile_relations_delta_set_to_vectors(__isl_take isl_set *set, void *user){
	dimInfo *info = (dimInfo*) user;
	stencil_scop* stencil = info->stencil;
	int dimPos = info->dimPos;

	if(dimPos < stencil->spaceInfoDim){
		char dimNamesList[256];
		isl_bool lower_bound, upper_bound;

		const char* dimName = isl_set_get_dim_name(set, isl_dim_set, dimPos);

		dimNamesList[0]='\0';
		{
			int i;
			for(i=0; i < stencil->spaceInfoDim; i++ ){
				sprintf(&dimNamesList[strlen(dimNamesList)], ",%s", isl_set_get_dim_name(set,isl_dim_set, i));
			}
		}
		dimNamesList[0] = ' ';

		{
			isl_set* tmpSet = isl_set_copy(set);

			if(dimPos > 0)
			{
				tmpSet = isl_set_remove_dims(tmpSet, isl_dim_set, 0, dimPos);
			}
			if(stencil->spaceInfoDim - (dimPos + 1) > 0)
			{
				tmpSet = isl_set_remove_dims(tmpSet, isl_dim_set, 1, stencil->spaceInfoDim - (dimPos + 1));
			}

			lower_bound = isl_set_dim_has_lower_bound(tmpSet, isl_dim_set, 0);
			upper_bound = isl_set_dim_has_upper_bound(tmpSet, isl_dim_set, 0);

			isl_set_free(tmpSet);
		}

		{
			char* minValStr = 0;
			char* maxValStr = 0;

			if(lower_bound == isl_bool_false && upper_bound == isl_bool_false){
				minValStr = malloc(2*sizeof(char));
				maxValStr = malloc(2*sizeof(char));

				sprintf(minValStr,"0");
				sprintf(maxValStr,"0");
			}
			else if(lower_bound == isl_bool_false){
				char buff[1024];
				maxValStr = get_pw_aff_value_to_str(isl_set_dim_max(isl_set_copy(set), dimPos));

				sprintf(buff,"{ [%s] : %s < 0 and %s < %s }", dimNamesList, dimName, dimName, maxValStr);

				minValStr = get_pw_aff_value_to_str(isl_set_dim_max(isl_set_intersect(isl_set_copy(set), isl_set_read_from_str(stencil->ctx, buff)), dimPos));
			}
			else if(upper_bound == isl_bool_false){
				char buff[1024];
				minValStr = get_pw_aff_value_to_str(isl_set_dim_min(isl_set_copy(set), dimPos));

				sprintf(buff,"{ [%s] : 0 < %s and %s < %s }", dimNamesList, dimName, minValStr, dimName);

				maxValStr = get_pw_aff_value_to_str(isl_set_dim_min(isl_set_intersect(isl_set_copy(set), isl_set_read_from_str(stencil->ctx, buff)), dimPos));
			}
			else{
				minValStr = get_pw_aff_value_to_str(isl_set_dim_min(isl_set_copy(set), dimPos));
				maxValStr = get_pw_aff_value_to_str(isl_set_dim_max(isl_set_copy(set), dimPos));
			}

			{
				char buff[1024];

				sprintf(buff,"{ [%s] : %s = %s or 0 = %s or %s = %s }", dimNamesList , minValStr, dimName, dimName, maxValStr, dimName );
				set = isl_set_intersect(set, isl_set_read_from_str(stencil->ctx, buff));
			}

			free(minValStr);
			free(maxValStr);
		}

		{
			dimInfo info;
			info.dimPos = dimPos + 1;
			info.stencil = stencil;

			isl_union_set *unionSet = isl_union_set_from_set(set);
			isl_union_set_foreach_set_in_set(unionSet, tile_relations_delta_set_to_vectors, &info);
			isl_union_set_free(unionSet);
		}
	}
	else{
		isl_debug_printf_set("\n#tile relatiions distance vectors set: %s\n", set);

		if(stencil->tileRelationsDeltaToScan)
			stencil->tileRelationsDeltaToScan = isl_union_set_union(stencil->tileRelationsDeltaToScan, isl_union_set_from_set(set));
		else
			stencil->tileRelationsDeltaToScan = isl_union_set_from_set(set);
	}

	return isl_stat_ok;
}

static isl_stat tile_relations_delta_set_point_analyze(__isl_take isl_point *pnt, void *user){
	int i;
	stencil_scop* stencil = (stencil_scop*) user;
	isl_set* set = isl_set_from_point(pnt);

	int dimIValue = get_pw_aff_value_to_int( isl_set_dim_max(isl_set_copy(set), 0 ));

	sprintf(&(stencil->tileWfSpaceEquations)[strlen(stencil->tileWfSpaceEquations)], " and %i*a1", dimIValue);

	for(i= 1; i < stencil->spaceInfoDim; i++){
		int dimIValue = get_pw_aff_value_to_int( isl_set_dim_max(isl_set_copy(set), i));

		sprintf(&(stencil->tileWfSpaceEquations)[strlen(stencil->tileWfSpaceEquations)], " + %i*a%i", dimIValue, i+1);
	}

	sprintf(&(stencil->tileWfSpaceEquations)[strlen(stencil->tileWfSpaceEquations)], " > 0");

	isl_set_free(set);

	return isl_stat_ok;
	return isl_stat_ok;
}

static __isl_give isl_set* tile_relations_delta_create_wafefront_hyperplane(stencil_scop* stencil){
	isl_set* ortogonalVector = 0;

	{
		int i;

		sprintf(stencil->tileWfSpaceEquations,"{[d0, b0, c0");
		for(i=0;i<stencil->spaceInfoDim;i++){
			sprintf(&(stencil->tileWfSpaceEquations)[strlen(stencil->tileWfSpaceEquations)], ", a%i", i + 1);
		}

		sprintf(&(stencil->tileWfSpaceEquations)[strlen(stencil->tileWfSpaceEquations)], "] : exists b1");
		for(i=1;i<stencil->spaceInfoDim;i++){
			sprintf(&(stencil->tileWfSpaceEquations)[strlen(stencil->tileWfSpaceEquations)], ", b%i", i + 1);
		}

		sprintf(&(stencil->tileWfSpaceEquations)[strlen(stencil->tileWfSpaceEquations)], " : ((b1 = a1 and a1 >= 0) or (b1 = -a1 and a1 < 0))");
		for(i=1;i<stencil->spaceInfoDim;i++){
			sprintf(&(stencil->tileWfSpaceEquations)[strlen(stencil->tileWfSpaceEquations)], " and ((b%i = a%i and a%i >= 0) or (b%i = -a%i and a%i < 0))", i + 1, i + 1, i + 1, i + 1, i + 1, i + 1);
		}

		sprintf(&(stencil->tileWfSpaceEquations)[strlen(stencil->tileWfSpaceEquations)], " and b0 > 0 and b0 = b1");
		for(i=1;i<stencil->spaceInfoDim;i++){
			sprintf(&(stencil->tileWfSpaceEquations)[strlen(stencil->tileWfSpaceEquations)], " + b%i", i + 1);
		}

		isl_union_set_foreach_point(stencil->tileRelationsDeltaToScan, tile_relations_delta_set_point_analyze, stencil);

		sprintf(&(stencil->tileWfSpaceEquations)[strlen(stencil->tileWfSpaceEquations)], " }");
	}


	{
		int i;
		char equationStr[4096];

		sprintf(equationStr,"{[d0 = 0, b0, c0");
		for(i=0;i<stencil->spaceInfoDim;i++){
			sprintf(&equationStr[strlen(equationStr)], ", a%i", i + 1);
		}

		sprintf(&equationStr[strlen(equationStr)], "] : exists c1");
		for(i=1;i<stencil->spaceInfoDim;i++){
			sprintf(&equationStr[strlen(equationStr)], ", c%i", i + 1);
		}

		sprintf(&equationStr[strlen(equationStr)], " : ((c1 = -1 and a1 >= 0) or (c1 = 0 and a1 < 0))");
		for(i=1;i<stencil->spaceInfoDim;i++){
			sprintf(&equationStr[strlen(equationStr)], " and ((c%i = -1 and a%i >= 0) or (c%i = 0 and a%i < 0))", i + 1, i + 1, i + 1, i + 1);
		}

		sprintf(&equationStr[strlen(equationStr)], " and c0 = c1");
		for(i=1;i<stencil->spaceInfoDim;i++){
			sprintf(&equationStr[strlen(equationStr)], " + c%i", i + 1);
		}

		for(i=0;i<stencil->spaceInfoDim;i++){
			sprintf(&equationStr[strlen(equationStr)], " and a%i != 0", i + 1);
		}


		sprintf(&equationStr[strlen(equationStr)], "}");

		{
			ortogonalVector = isl_set_lexmin(isl_set_intersect(isl_set_read_from_str(stencil->ctx, stencil->tileWfSpaceEquations), isl_set_read_from_str(stencil->ctx, equationStr)));

			isl_debug_printf("\n#wafefront hyperplane: lexmin( %s * ", equationStr);
			isl_debug_printf("%s )\n", stencil->tileWfSpaceEquations);

			isl_debug_printf_set("\n#result: %s\n", ortogonalVector);
		}
	}

	return ortogonalVector;
}

static void stencil_global_schedule_create(__isl_keep isl_union_set *domain_union_set, __isl_keep isl_union_set *orig_domain_union_set, __isl_keep isl_union_map *mapper_union_map, stencil_scop* stencil){

	isl_debug_printf("%s", "\n#create stencil schedule code start\n");

	isl_debug_printf_union_map("\n#stencilScheduleMap: %s\n", stencil->stencilScheduleMap);
	isl_debug_printf_union_set("\n#domain_union_set: %s\n", domain_union_set);

	stencil->stencilGlobalSchedule = isl_union_map_intersect_domain(isl_union_map_copy(stencil->stencilScheduleMap), isl_union_set_copy(domain_union_set));
	isl_debug_printf_union_map("\n#stencilGlobalSchedule: %s\n", stencil->stencilGlobalSchedule);
	isl_debug_printf_union_map("\n#mapper_union_map: %s\n", mapper_union_map);

	stencil->stencilGlobalSchedule = isl_union_map_apply_range(isl_union_map_copy(mapper_union_map),stencil->stencilGlobalSchedule);
	stencil->stencilGlobalSchedule = isl_union_map_intersect_domain(stencil->stencilGlobalSchedule, isl_union_set_copy(orig_domain_union_set));
	stencil->stencilGlobalSchedule = isl_union_map_coalesce(isl_union_map_remove_divs(stencil->stencilGlobalSchedule));
	isl_debug_printf_union_map("\n#stencilGlobalSchedule: %s\n", stencil->stencilGlobalSchedule);

	isl_debug_printf("%s", "\n#create stencil schedule code end\n");
}

static void stencil_global_schedule_scalar_create(__isl_keep isl_union_map *schedule_map, stencil_scop* stencil, calc_scop* calc){
	int i;
	char bufWfStr[256];
	char* strScheduleMap = isl_union_map_to_str(schedule_map);

	bufWfStr[0] = '\0';
	sprintf(&bufWfStr[strlen(bufWfStr)],"] -> [%i, %s, ", stencil->stencil_index, "0");
	for(i = 0; i < calc->scopeDimSize; i++){
		sprintf(&bufWfStr[strlen(bufWfStr)],"%s, ", "0");
	}
	strScheduleMap = str_replace(strScheduleMap, "] -> [", bufWfStr);
	strScheduleMap = str_replace(strScheduleMap, ", ]", "]");
	stencil->stencilScheduleMap = isl_union_map_read_from_str(stencil->ctx, strScheduleMap);

	free(strScheduleMap);


	stencil->stencilGlobalSchedule = isl_union_map_intersect_domain(isl_union_map_copy(stencil->stencilScheduleMap), isl_union_set_copy(calc->domain));
	stencil->stencilGlobalSchedule = isl_union_map_apply_range(isl_union_map_copy(calc->schedule_mapper),stencil->stencilGlobalSchedule);
	stencil->stencilGlobalSchedule = isl_union_map_intersect_domain(stencil->stencilGlobalSchedule, isl_union_set_copy(calc->orig_domain));
	stencil->stencilGlobalSchedule = isl_union_map_coalesce(isl_union_map_remove_divs(stencil->stencilGlobalSchedule));
}

static void stencil_global_schedule_card_calculate(stencil_scop* stencil){
	isl_debug_printf("\n%s\n", "#wavefront card calculate start");

	isl_union_set_foreach_set_in_set(stencil->tile, stencil_global_schedule_card_set_build,stencil);
	isl_union_set_foreach_set_in_set(stencil->tile, stencil_global_schedule_card_union_set_build,stencil);

	stencil->stencilScheduleCardSet = isl_set_remove_divs(stencil->stencilScheduleCardSet);
	stencil->stencilScheduleCardUnionSet = isl_union_set_remove_divs(stencil->stencilScheduleCardUnionSet);

	isl_debug_printf_set("\n#wavefront card set: %s\n", stencil->stencilScheduleCardSet);
	isl_debug_printf_union_set("\n#wavefront card union set: %s\n", stencil->stencilScheduleCardUnionSet);

	stencil->wavefrontSetCardValue = isl_pw_qpolynomial_max(isl_set_card(isl_set_copy(stencil->stencilScheduleCardSet)));

	isl_debug_printf("\n%s\n", "#Wavefront card calculate end");
}

static isl_stat stencil_global_schedule_card_set_build(__isl_take isl_set *set, void *user){
	stencil_scop* stencil = (stencil_scop*)user;
	char* setStr = 0;
	char bufStr[4096];

	bufStr[0]= '\0';

	{
		isl_basic_set* basicSet = 0;
		set = isl_set_set_tuple_name(set, "");
		setStr= isl_set_to_str(set);
		isl_set_free(set);

		basicSet = isl_basic_set_read_from_str(stencil->ctx, setStr);
		free(setStr);
		setStr = isl_basic_set_to_str(basicSet);
		isl_basic_set_free(basicSet);
	}

	{
		char* paramsList = str_substring(setStr, "[", "] -> {");
		char* indexesList = str_substring(setStr, "{ [", "]");
		char* condition = str_substring(setStr, ":", "}");

		sprintf(&bufStr[strlen(bufStr)], "[] -> { [%s] : exists %s, %s :", stencil->wfIndex, paramsList, indexesList);
		if(strlen(stencil->domainParams)>0){
			sprintf(&bufStr[strlen(bufStr)], " %s ", stencil->domainParamsBounds);
		}
		sprintf(&bufStr[strlen(bufStr)], " %s %s }", stencil->tileWfSpaceExpr, condition);

		free(condition);
		free(indexesList);
		free(paramsList);
	}

	{
		isl_set* stencilScheduleCardSet = isl_set_read_from_str(stencil->ctx, bufStr);
		isl_debug_printf("\n#card set: %s \n", bufStr);
		isl_debug_printf_set("\n#result: %s\n", stencilScheduleCardSet);

		if(!stencil->stencilScheduleCardSet){
			stencil->stencilScheduleCardSet = isl_set_copy(stencilScheduleCardSet);
		}
		stencil->stencilScheduleCardSet = isl_set_coalesce(isl_set_union(stencil->stencilScheduleCardSet, stencilScheduleCardSet));
	}

	free(setStr);
	return isl_stat_ok;
}

static isl_stat stencil_global_schedule_card_union_set_build(__isl_take isl_set *set, void *user){
	stencil_scop* stencil = (stencil_scop*)user;
	char* setStr = 0;
	char bufStr[4096];

	bufStr[0]= '\0';

	{
		isl_basic_set* basicSet = 0;
		set = isl_set_set_tuple_name(set, "");
		setStr= isl_set_to_str(set);
		isl_set_free(set);

		basicSet = isl_basic_set_read_from_str(stencil->ctx, setStr);
		free(setStr);
		setStr = isl_basic_set_to_str(basicSet);
		isl_basic_set_free(basicSet);
	}

	{
		int i;
		char spaceIndexList[256];
		char* indexesList = str_substring(setStr, "{ [", "]");
		char* condition = str_substring(setStr, ":", "}");

		sprintf(spaceIndexList,"%s", stencil->spaceInfo[0].spaceName);
		for(i=1; i<stencil->spaceInfoDim;i++){
			sprintf(&spaceIndexList[strlen(spaceIndexList)],", %s", stencil->spaceInfo[i].spaceName);
		}

		sprintf(&bufStr[strlen(bufStr)], "[%s] -> { [%s] : exists %s, %s :", stencil->domainParams, stencil->wfIndex, spaceIndexList, indexesList);
		sprintf(&bufStr[strlen(bufStr)], " %s %s }", stencil->tileWfSpaceExpr, condition);

		free(condition);
		free(indexesList);
	}

	{
		isl_union_set* stencilScheduleCardUnionSet = isl_union_set_read_from_str(stencil->ctx, bufStr);
		isl_debug_printf("\n#card union set: %s \n", bufStr);
		isl_debug_printf_union_set("\n#result: %s\n", stencilScheduleCardUnionSet);

		if(!stencil->stencilScheduleCardUnionSet){
			stencil->stencilScheduleCardUnionSet = isl_union_set_copy(stencilScheduleCardUnionSet);
		}
		stencil->stencilScheduleCardUnionSet = isl_union_set_coalesce(isl_union_set_union(stencil->stencilScheduleCardUnionSet, stencilScheduleCardUnionSet));
	}

	free(setStr);
	return isl_stat_ok;
}

