/*
 * stencil_scop.h
 *
 *  Created on: 22 pa 2018
 *      Author: Maciej
 */
#include <isl/ctx.h>
#include <isl/union_map.h>
#include "calc_scop.h"
#include <barvinok/isl.h>

#ifndef STENCIL_SCOP_H_
#define STENCIL_SCOP_H_


#if defined(__cplusplus)
extern "C" {
#endif

typedef struct stencil_Param{
	int size;
	int sizesDim;
	int sizesDimIndex;
	int *sizes;
	int method;
	int bound;
} stencil_Param;


typedef struct deltaDim_Info{
	isl_bool isNoUpperBound;
	isl_bool isNoLowerBound;
	float positive_val;
	char* positive_vector_set;
	float negative_val;
	char* negative_vector_set;
} deltaDim_Info;

typedef struct space_Info{
	char* spaceName;
	int* vector;
} space_Info;

typedef struct hyperplane_check_info{
	int lastId;
	char paramIdList[256];
	char paramIdCondition[256];
	char** statementList;
} hyperplane_check_info;

typedef struct stencil_scop{

	isl_ctx *ctx;
	int stencil_index;

	int loopDim;
	const char (*loopDimName)[8];
	char* domainParams;
	char* domainParamsBounds;

	isl_union_set* delta_union_set_to_scan;
	deltaDim_Info* deltaDimInfo;
	isl_union_set* deltaInfoSet;

	isl_bool isNegParam;

	stencil_Param* stencilParam;
	int spaceInfoDim;
	char spaceInfoEquations[4096];
	char spaceInfoEquationsLexico[4096];
	isl_set* spaceInfoHyperplanes;
	int* spaceInfoHyperplanesDims;
	space_Info* spaceInfo;
	isl_set* spaceInfoSet;


	isl_union_set* tile;
	isl_union_set* tileOrder;
	isl_union_map* tileRelations;
	isl_union_set* tileRelationsDelta;
	isl_union_set* tileRelationsDeltaToScan;
	char tileWfSpaceEquations[4096];
	char tileWfSpaceExpr[256];
	char wfIndex[8];
	isl_union_map* tileScheduleMap;

	isl_union_map* stencilScheduleMap;
	isl_union_map* stencilGlobalSchedule;

	hyperplane_check_info* hyperplaneCheckInfo;

	isl_set* stencilScheduleCardSet;
	isl_union_set* stencilScheduleCardUnionSet;

	isl_val* wavefrontSetCardValue;
} stencil_scop;

typedef struct stencil_scop_list{
	int counter;
	stencil_scop **stencil_scop_list;
	isl_union_map *stencilGlobalSchedule;
} stencil_scop_list;

__isl_give stencil_scop_list* stencil_scop_list_from_calc_scop_list(isl_ctx *ctx, __isl_keep calc_scop_list* calc_list, stencil_Param* stencilParam);
__isl_null stencil_scop_list* stencil_scop_list_free(stencil_scop_list* stencil_list);

__isl_give stencil_scop* stencil_scop_from_calc_scop(isl_ctx *ctx, __isl_keep calc_scop* calc, stencil_Param* stencilParam, int stencil_index);
__isl_give stencil_scop* stencil_scop_alloc(isl_ctx *ctx);
__isl_null stencil_scop *stencil_scop_free(__isl_take stencil_scop *stencil);
char* stencil_scop_deltaDim_Info_to_str(__isl_keep stencil_scop *stencil, int pos);
char* stencil_scop_space_Info_to_str(__isl_keep stencil_scop *stencil, int pos);

#if defined(__cplusplus)
}
#endif

#endif /* STENCIL_SCOP_H_ */
